use std::{
	env, fs::OpenOptions, ops::Deref, include_str,
};
use memmap2::MmapOptions;
use rouille::{Response, router};
use btrfs_explorer::main_error::MainError;
use btrfs_explorer::render_common::http_path_set;

const CSS_FILE: &'static str = include_str!("style.css");

fn main() -> Result<(), MainError> {
	let args: Vec<String> = env::args().collect();

	if args.len() < 2 {
		return Err("Argument required".into());
	}

	let filename: &str = args[1].as_ref();
	let sockaddr: &str = args.get(2)
		.map_or("localhost:8080", Deref::deref);
	let http_path: String = args.get(3)
		.map_or(String::new(), Clone::clone);
	http_path_set(http_path);

	let file = OpenOptions::new().read(true).open(filename)?;
	let image = unsafe { MmapOptions::new().map(&file)? };

	rouille::start_server(sockaddr, move |request| {
		println!("Request: {}", request.url());

		router!(
			request,
			(GET) ["/"] =>
				btrfs_explorer::http_chunk::http_allchunks(&image, request).unwrap(),
			(GET) ["/root"] =>
				btrfs_explorer::http_tree::http_root(&image, None, request),
			(GET) ["/chunk/{offset}", offset: String] =>
				btrfs_explorer::http_chunk::http_chunk(&image, &offset, request).unwrap(),
			(GET) ["/tree/{tree}", tree: String] =>
				btrfs_explorer::http_tree::http_tree(&image, &tree, None, request.get_param("key").as_deref(), request).unwrap(),
			(GET) ["/tree/{tree}/{key}", tree: String, key: String] =>
				btrfs_explorer::http_tree::http_tree(&image, &tree, None, Some(&key), request).unwrap(),
			(GET) ["/tree/{tree}?key={key}", tree: String, key: String] =>
				btrfs_explorer::http_tree::http_tree(&image, &tree, None, Some(&key), request).unwrap(),
			(GET) ["/tree/{tree}/{method}/{key}", tree: String, method: String, key: String] =>
				btrfs_explorer::http_tree::http_tree(&image, &tree, Some(&method), Some(&key), request).unwrap(),
			(GET) ["/favicon.ico"] => Response::empty_404(),
			(GET) ["/style.css"] => Response::from_data("text/css", CSS_FILE)
				.with_additional_header("Cache-Control", "max-age=3600"),
			_ => Response::empty_404(),
		)
	});
}
